/* <a href="http://www.ooRexx.org>Object Rexx</a> convenience class for making IBM's and
   Apache's Bean Scripting Framework (BSF) available in an OO-style as well: Java classes and
   Java objects appear as Object Rexx classes and objects and one can send Object Rexx
   messages to them.

   Object Rexx programmers should appreciate this little class ...
   ;)

   date:
          2005-06-05, rgf: - renamed BSF-pass through class methods to begin with "BSF.",
                           - using pre-registered Java-class proxies for box() and unbox(),
                           changed bsf.import logic to allow for dynamic allocating (if a
                           Java object proxy needs to access its Java class object):
                           - added instance method "bsf.class" to return the Java class object
                           proxy, which allows accessing the static methods and fields as well;


          2005-05-07, rgf: added bsf.getFieldValueStrict (strict case!),
                      methods having a trailing "=" are interpreted as setter methods, causing
                      the "setFieldValue" subfunction to be invoked using the method name without
                      the trailing equal sign as the name of the field

          2005-05-02, rgf, imported the primitive type wrapper classes into ooRexx
                           added box() and unbox(),
                           .box.true and .box.false (boxed as Booleans): -- removed 2005-05-03, rgf

          2004-03-04, rgf, test if BSFLoadJava() is registered before invoking it
          2004-02-22, rgf, trap exception on registering external Rexx-functions; needed
                      for Tobias Specht for debugging BWS on Konqueror;
                      partially removing a reported bug by TS (ORX not able handling literals
                      larger than 250 chars in INTERPRET)

          2003-11-19, rgf, added "RAISE PROPAGATE" statements in order for Rexx programs
                      being able to intercept exceptions which may occur in programs using
                      the proxy classes of this file

          2003-09-05, rgf, added "newStrict" floating constructor method for imported
                      Java classes and "addNewStrict" class method to attach it to
                      Object Rexx proxy classes by the "import" class method (i.e. setting
                      up an Object Rexx class as a proxy), to allow using strictly typed
                      arguments at instantiation time

          2003-06-01, rgf, added Apache license

          2003-05-10, rgf, changed "_new" to "Strict", changed arrays to never
                      use strict typing (not necessary as type can be inferred
                      from array itself)

          2003-05-07, rgf, using the new "invoke_new", which allows for not
                      using type information with the arguments

          2003-03-10, rgf, changed ".bsf.cls" to ".bsf4rexx" and ".JC" to ".class"
          2003-02-09, rgf, no dimensional limits on arrays anymore,
                           makearray(), supplier() now created at the Java side, merely re-using

          2003-01-29, Augsburg-->Vienna (train), adding direct JVM-loading support,
                      BSF_PROXY, .BSF.cls (for accessing pre-registered Java class objects)

          2003-01-28, Augsburg, renamed "BSF_ARRAY" to "BSF_ARRAY_PROXY" to denote its purpose
          2001-04-22, 2001-05-02 (on the airplane, returning home
                                  from the 12th International Rexx Symposium),
          2001-05-07 (fixed bug in bsf:init, if directly used to instantiate data)
          2001-05-26 (made method "bsf.exit" available at the instance level)
          2001-06-02 (BSF: made method "bsf.sleep" available at the instance level and
                      as "sleep" at the class level;
                      BSF_ARRAY_PROXY: MAKEARRAY and SUPPLIER now only return entries with
                                 values as does Object Rexx with native arrays)

   version:     1.3.1

   authors:     (c) 2001-2003 Rony G. Flatscher, University of Augsburg, University of Essen,
                    WU (pronounced: "vey-uh") University Vienna
                         Peter Kalender (October 2000 - February 2001), University of Essen, feasibility study

  If this program is part of a distribution of an Apache foundation project (e.g. BSF,
  the Bean Scripting Framework), then the immediately following Apache foundation license
  applies, else the LGPL license which follows thereafter. (Should you need another license
  please contact Rony G. Flatscher at (rony.flatscher@wu-wien.ac.at>.)

Apache foundation license:

/*
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "Apache BSF", "Apache", and "Apache Software Foundation"
 *    must not be used to endorse or promote products derived from
 *    this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions on behalf of the Apache
 * Software Foundation and was originally created by* Rony G. Flatscher.
 * For more information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */




LGPL license:

 /*  LICENSE:
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
  *  License as published by the Free Software Foundation; either
  *  version 2.1 of the License, or (at your option) any later version.
  *
  *  This library is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  *  Lesser General Public License for more details.
  *
  *  You should have received a copy of the GNU Lesser General Public
  *  License along with this library; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
  *  The following URL may be available for further information:
  *
  *     <a href="http://www.gnu.org/copyleft/lesser.html">
  *              http://www.gnu.org/copyleft/lesser.html
  *     </a>
  */
*/

      -- load the BSF4Rexx functions and start a JVM, if necessary
   if rxFuncQuery("BSF") = 1 then   -- BSF() support not registered yet ?
   do
      call registerBSFFunctions  -- register  BSF-functions, load JVM
   end

   -- ---rgf, initialization, make the pre-registered class objects accessible via
   -- ".bsf4rexx": beware using these classes without studying their definition and purpose !

   .local~bsf4rexx=.directory~new
   .local~b4r=.bsf4rexx             -- create shorter alias for .BSF4Rexx-directory

   arr=.array~of( ,
      "Class.class",      'Class.class'        ,,   -- a Java Object class
      "Object.class",     'Object.class'       ,,   -- a Java Object class
      "Method.class",     'Method.class'       ,,   -- a Java Object class
       ,
      "Array.class",      'Array.class'        ,,   -- a Java Object class
      "String.class",     'String.class'       ,,   -- a Java Object class
      "System.class",     'System.class'       ,,   -- a Java Object class
       ,
      "boolean",          'boolean.class'      ,,   -- primitive
      "boolean.class",    'Boolean.class'      ,,   -- a Java Object class
      "byte",             'byte.class'         ,,   -- primitive
      "byte.class",       'Byte.class'         ,,   -- a Java Object class
      "char",             'char.class'         ,,   -- primitive
      "Character.class" , 'Character.class'    ,,   -- a Java Object class
      "double",           'double.class'       ,,   -- primitive
      "Double.class",     'Double.class'       ,,   -- a Java Object class
      "float",            'float.class'        ,,   -- primitive
      "Float.class",      'Float.class'        ,,   -- a Java Object class
      "int",              'int.class'          ,,   -- primitive
      "Integer.class",    'Integer.class'      ,,   -- a Java Object class
      "long",             'long.class'         ,,   -- primitive
      "Long.class",       'Long.class'         ,,   -- a Java Object class
      "short",            'short.class'        ,,   -- primitive
      "Short.class",      'Short.class'        ,,   -- a Java Object class
      "void",             'void.class'         ,,   -- primitive
      "Void.class",       'Void.class'         )    -- a Java Object class

  do i=1 to arr~items by 2
     .bsf4rexx~setentry(arr[i], .bsf_reference~new(arr[i+1]))
  end

      -- Java classes representing the primitive datatypes
   arr=.array~of( ,
      'java.lang.String'            ,,   -- a Java Object class
      'java.lang.Boolean'           ,,   -- a Java Object class
      'java.lang.Byte'              ,,   -- a Java Object class
      'java.lang.Character'         ,,   -- a Java Object class
      'java.lang.Double'            ,,   -- a Java Object class
      'java.lang.Float'             ,,   -- a Java Object class
      'java.lang.Integer'           ,,   -- a Java Object class
      'java.lang.Long'              ,,   -- a Java Object class
      'java.lang.Short'             )    -- a Java Object class

  do i=1 to arr~items
     .bsf~bsf.import(arr[i], arr[i])  -- make these classes accessible as if they werer ooRexx classes
  end


-- box primitive datatypes into their Java classes counterparts
-- BOolean, BYte, C, D, F, I, L, SHort, STring

:: routine bsf.box  public
   parse upper arg type +2, val

   select
     when abbrev(type, "BO") then
          do
             if val=.true | val~left(1)="T" then val="true"
                                            else val="false"
             return .java.lang.Boolean~new(val)
          end

     when abbrev(type, "BY") then return .java.lang.Byte~new(val)

     when abbrev(type, "C")  then return .java.lang.Character~new(arg(2)) -- use val verbatimly (retain mixed case)
     when abbrev(type, "D")  then return .java.lang.Double~new(val)

     when abbrev(type, "F")  then return .java.lang.Float~new(val)
     when abbrev(type, "I")  then return .java.lang.Integer~new(val)
     when abbrev(type, "L")  then return .java.lang.Long~new(val)
     when abbrev(type, "SH") then return .java.lang.Short~new(val)
     when abbrev(type, "ST") then return .java.lang.String~new(arg(2))    -- use val verbatimly (retain mixed case)

     otherwise return arg(2)
   end

-- unbox primitive datatype from their Java class counterpart, where they are embedded
-- BOolean, BYte, C, D, F, I, L, SHort, STring
:: routine bsf.unbox public
   use arg o

     -- extract class name from proxy's objectName (e.g. "java.lang.String@a3b0cdef")
   parse upper value o~objectname with . "." . "." className "@" .

     -- if not a wrapper class return object unchanged
   if pos(className, "BOOLEAN BYTE CHARACTER DOUBLE FLOAT INTEGER LONG SHORT STRING")=0 then return o

     -- toString() would return instead of "1"/"0" the values "true"/"false"
   if abbrev(className, "BO")  then return o~booleanValue

   return o~toString -- Java string represantation suffices for ooRexx



-- register BSF() external functions, assume that Java needs to be loaded
:: routine registerBSFFunctions

   call rxFuncAdd "BsfLoadFuncs", "BSF4Rexx", "BsfLoadFuncs"
   signal on syntax -- name weiter
   call BsfLoadFuncs    -- register all the BSF4Rexx functions ...
syntax:
   -- signal off syntax
   if var("SIGL") then
   do
      say "BSF.CLS: SIGL="sigl "= line # with exception, line: ["sourceline(sigl)"]" -- nur bei exception
      say "         exception:" condition("C")
      say "         traceback:"
      do line over condition("o")~traceback
          say "                  " line
      end
   end

   if rxFuncQuery("BSFLOADJAVA") = 0 then -- load JVM only, if BSFLoadJava() is registered
   do
      -- .error~say( "... loading JVM ...")
      call BsfLoadJava     -- load JVM
   end
   else
   do
      .error~say("... cannot load the JVM, BSFLoadJava() not registered.")
   end

   return

/* -------------------------------------------------------------
   routine which turns an array into an argument string as expected
   by the Java side, starting with the indicated element
   ------------------------------------------------------------- */
:: ROUTINE makeArgString
   use arg args

   tmpString    = ""
   do i=1 to args~items
      if args[i]=.nil then
         tmpString = tmpString ', "' || .bsf~bsf.null_string || '"'
      else
         tmpString = tmpString ', "' || args[i] || '"'
   END
   return tmpString


/* -------------------------------------------------------------
   routine which mimicks C/Java's "x ? a : b;"
   if arg(1) is true, returns second argument, else third
   ------------------------------------------------------------- */
:: ROUTINE choose
   if arg(1)=.true then return arg(2)
                   else return arg(3)


/* -------------------------------------------------------------
   floating class method meant for attaching to imported Java
   classes (for which an Object Rexx proxy class is created)
   to construct an instance (a "bean") with strict arguments (i.e.
   type information are given for each argument; arguments are
   passed to BSF which uses them to create the bean instance
   from the given Java class, ---rgf, 2003-09-05
   ------------------------------------------------------------- */
::METHOD newStrict
  USE ARG javaClass, args

  instance=.bsf_reference~new("uninitialized BSF-object!")

        -- create an instance of javaClass and register it with BSF,
        -- leave beanName empty, this way we get a unique name from Java
/* 2004-02-22
  if args~class = .array then tmpStr=makeArgString(args)
                         else tmpSTr=makeArgString(ARG(1, "Array"))

  code = 'beanName=BSF( "registerBeanStrict", , "' || self~java_path || '"' tmpStr ")"
*/

   if args~class = .array then tmpArr=args            -- array of arguments
                          else tmpArr=arg(1, "Array") -- create an array from arguments

   code = 'beanName=BSF( "registerBeanStrict", , "' || self~bsf.FullJavaClassName || '"'
   do i=1 to tmpArr~items
      a.i=choose(tmpArr[i]=.nil, .bsf~bsf.null_string, tmpArr[i])
      code=code ", a."i
   end
   code=code ")"

  -- say "newStrict, code: ["code"]..."


  signal on any         -- trap any exceptions
  interpret code        /* execute the code, calling bsf() with 'registerBean'  */

        -- assign beanName to object, use the Java supplied name
  instance~ObjectName=beanName

  return instance

any: raise propagate  -- raise the exception in caller/invoker on the Rexx side



/* -------------------------------------------------------------
   routine which checks whether data returned from the Java side
   represents a Java object; if so, an Object Rexx proxy is created
   and returned, else the received string value
   ------------------------------------------------------------- */
:: ROUTINE bsf.checkResult public
   parse arg data

   if data=.bsf~bsf.null_string then return .nil

        /* if an object was returned, it has been registered on the Java side;
           therefore reuse, if possible or, create an Object Rexx proxy of it;
           first check whether the name was constructed on the Java side        */
  if pos("@", data)>0 then      -- is this a beanName, created by Java?
  do
     parse var data before '@' after
     if before <> "" then
     do
        if datatype(after, "X") then    -- o.k. second part a hexadecimal value?
        do
           if pos(" ", after)=0 then    -- o.k. no space there?
           do
              -- create an Object Rexx proxy object for the Java object

              dimension = verify(data, "[")-1  -- is it a Java array?
              -- if dimension>0 & dimension<6 then
              if dimension>0 then   -- rgf, 2003-02-09, no upper limit anymore
                 return .bsf_array_reference~new(data, dimension) -- return a BSF_ARRAY proxy

              return .bsf_reference~new(data)   -- return a BSF_REFERENCE prox
           end
        end
     end
  end

  return data           -- assume the string of a primitive type




/* *************************************************************
   Class:       IBM's "Bean Scripting Framework"
   Purpose:     the proxy super-class for the Java classes
   ************************************************************* */
::CLASS BSF PUBLIC

/* -------------------------------------------------------------
   class method which merely initializes the class attribute "bsf.null_string"
   to its default value
   ------------------------------------------------------------- */
::METHOD init           CLASS
  EXPOSE bsf.null_string bsf.FullJavaClassName bsf.JavaClassObjectProxies bsf.JavaClassObjects
  bsf.null_string=".NIL"
  bsf.FullJavaClassName=.nil     -- this is an Object Rexx path, hence no fully
                         -- qualified Java class available

  bsf.JavaClassObjectProxies=.directory~new
  bsf.JavaClassObjects      =.directory~new


/* -------------------------------------------------------------
   class method to allow invoking static methods via ooRexx
   ------------------------------------------------------------- */
::method unknown class
-- say self~string"::unknown" "self~bsf.FullJavaClassName="self~bsf.FullJavaClassName

  signal on syntax name any
  -- if bsf.getJavaClassObject[self~bsf.FullJavaClassName]=.nil then   -- no fully qualified Java class ("path") given ?
  if self~bsf.FullJavaClassName=.nil then   -- no fully qualified Java class ("path") given ?
  do
      -- Object "object" does not understand message "message"
     raise syntax 97.1 array(self~string, arg(1))
  end

     -- forward unknown message to the BSF proxy object representing the Java class object,
     -- this way it becomes possible to access the Java static (class) methods quite easily

  forward to (.bsf~bsf.getJavaClassObjectProxy(self~bsf.FullJavaClassName))

  any: raise propagate              -- raise exception in caller


/* -------------------------------------------------------------
   class method allows for creating an ooRexx proxy class for a Java class
   which is made available to this process by placing it (optionally) into the
   .local environment
   ------------------------------------------------------------- */
::METHOD bsf.import class
--  USE ARG classname, bsf.FullJavaClassName
  USE ARG java_path, classname   -- 2005-06-05, changed argument sequence

  bNoClassName=arg(2, "Omitted")

  signal on syntax name any

-- say "... java_path="java_path "className="classname "..."

  o=.bsf~bsf.JavaClassObjects[java_path] -- try to get the ooRexx class object proxy
  -- o=bsf.JavaClassObjectProxies[java_path]  -- do we have the Java class imported already, if so return proxy class object

  if o<>.nil then return o

      -- let Java create the class object, which we can refer to
  jco=.bsf4rexx~class.class~forName(java_path)  -- get Java class object (can be a BSF_Reference !)
  .bsf~bsf.JavaClassObjectProxies[java_path]=jco     -- save jco in directory

-- say "/// java_path=["java_path"]" "self="self
      -- now create proxy class object

  if bNoClassName then
     new_class = .BSF~subclass(java_path) -- create a subclass of .BSFClass (a proxy ooRexx class)
  else
     new_class = .BSF~subclass(classname) -- create a subclass of .BSFClass (a proxy ooRexx class)

  .bsf~bsf.JavaClassObjects[java_path]=new_class    -- save the ooRexx Java class object proxy

      -- attach a "newStrict" constructor method to the Object Rexx class object
  new_class~attachNewStrict( java_path )

      -- create code for the INIT-method of the new class
  method_txt = 'self~init:super("' || java_path || '", ARG(1,"A"))'
  new_class~define('init', method_txt)  -- define the constructor instance method "INIT"

--  if bNoClassName | arg(2)=.nil then say "... about returning:" new_class
  if bNoClassName | arg(2)=.nil then return new_class

  -- save the new BSF proxy class in .local to make it available as any other ooRexx class
  .local~setentry(classname, new_class) -- put new ooRexx proxy class into the local environment

-- say ">>> java_path="java_path "className="className "about returning:" new_class
  return new_class

  any: raise propagate              -- raise exception in caller


/* -------------------------------------------------------------
   allows to retrieve the fully qualified Java class name, if a
   Java class got imported (encapsulated in an Object Rexx class
   proxy); needs to be a method in order to be able to use Object's
   SETMETHOD
   ------------------------------------------------------------- */
::METHOD attachNewStrict class
  use arg java_path
  if self~superclasses[1]<>.bsf then return   -- not a direct subclass of "BSF",
                                              -- hence not from "IMPORT"
  self~setmethod("newStrict", .methods["NEWSTRICT"])
  self~bsf.FullJavaClassName=java_path    -- save fully qualified Java class name


/* -------------------------------------------------------------
   allows to retrieve the fully qualified Java class name, if a
   Java class got imported (encapsulated in an Object Rexx class
   proxy); not directly accessible from outside the class
   ------------------------------------------------------------- */
::method bsf.FullJavaClassName class  -- can be read from any object
  expose bsf.FullJavaClassName
  return bsf.FullJavaClassName

::method  "bsf.FullJavaClassName=" class private  -- can only be set via "self"
  expose  bsf.FullJavaClassName
  use arg bsf.FullJavaClassName


/* -------------------------------------------------------------
   class method which returns the BSF proxy object wrapping
   a Java class object
   ------------------------------------------------------------- */
::method bsf.getJavaClassObjectProxy class
  use arg javaClassName

  o=self~bsf.JavaClassObjectProxies[javaClassName]   -- try to get proxy for Java class object
  if o=.nil then           -- no proxy for Java class object yet, create one
  do
     .bsf~bsf.import(javaClassName) -- import the Java class
     o=self~bsf.JavaClassObjectProxies[javaClassName]   -- try to get proxy for Java class object
  end
  return o        -- return the BSF Java class proxy



/* -------------------------------------------------------------
   directory of BSF proxy objects referring to the Java class object
   ------------------------------------------------------------- */
::method bsf.JavaClassObjectProxies  ATTRIBUTE CLASS

/* -------------------------------------------------------------
   directory of ooRexx BSF proxy objects classes having new() and
   newStrict() defined on them
   ------------------------------------------------------------- */
::method bsf.JavaClassObjects        ATTRIBUTE CLASS

/* -------------------------------------------------------------
   class attribute method stores and allows access to the string
   which is used to indicat "null" for the Java side; upon return
   of values from Java, a Java "null" will be represented via
   Object Rexx' .nil object
   ------------------------------------------------------------- */
::METHOD bsf.null_string    ATTRIBUTE CLASS



/* -------------------------------------------------------------
   method to construct an instance (a "bean"); arguments are
   passed to BSF which uses them to create the bean instance
   from the given Java class
   ------------------------------------------------------------- */
::METHOD init
  USE ARG javaClass, args

        -- create an instance of javaClass and register it with BSF,
        -- leave beanName empty, this way we get a unique name from Java
--  code = 'beanName=BSF( "registerBean", , "' || javaClass || '"' makeArgString(ARG(2, "Array")) ")"

  /* 2004-02-22
  if args~class = .array then tmpStr=makeArgString(args)
                         else tmpSTr=makeArgString(ARG(2, "Array"))

--  code = 'beanName=BSF( "registerBean", , "' || javaClass || '"' makeArgString(args) ")"
  code = 'beanName=BSF( "registerBean", , "' || javaClass || '"' tmpStr ")"
  */


   if args~class = .array then tmpArr=args            -- array of arguments
                          else tmpArr=arg(2, "Array") -- create an array from arguments

   code = 'beanName=BSF( "registerBean", , "' || javaClass || '"'
   do i=1 to tmpArr~items
      a.i=choose(tmpArr[i]=.nil, .bsf~bsf.null_string, tmpArr[i])
      code=code ", a."i
   end
   code=code ")"

  -- say "init, code: ["code"]..."

  signal on any         -- trap any exceptions
  interpret code        /* execute the code, calling bsf() with 'registerBean'  */

        -- assign beanName to object, use the Java supplied name
  self~ObjectName=beanName

-- say "---> BSF.init  : bean=["||beanName||"]"  -- 2003-01-05, ---rgf
  return self

any: raise propagate  -- raise the exception in caller/invoker on the Rexx side

/* -------------------------------------------------------------
   destructor method (called by Object Rexx when object gets
   destroyed) to unregister the bean from the BSF registry; this
   then allows the JVM to garbage collect that Java object as well
   ------------------------------------------------------------- */
::METHOD uninit
-- say "---> BSF.uninit: bean=["||self~objectname||"]" -- 2003-01-05, ---rgf

  return BSF("unregisterBean", self~objectname)


/* -------------------------------------------------------------
   unknown method (gets called by Object Rexx whenever an unknown
   message was sent to an instance of this class); forwards the
   message to BSF including all of the received arguments
   ------------------------------------------------------------- */
::METHOD unknown
  USE ARG name, args

  signal on syntax name any         -- trap any exceptions

-- say " unknown(): name=["name"], arg()="arg() "arg(1)="arg(1)

-- say "arg(2)="arg(2)~string
if args~class=.array then
do
--    say "            args~items="args~items
--   if args~items>0 then say "            args[1]="args[1]

--    say ' right(name,1)="=" ' (right(name,1)="=")
     if (right(name,1)="=") then -- field assignment?
     do
--    say "BINGO!"
        items=args~items
        parse var name tmpName "=" . -- extract message name without '='
--    say "... tmpName=["tmpName"]"
        if items=1 then             -- arg(1)=value to assign to field
           return self~bsf.setFieldValue(tmpName, args[1])
        else if items=2 then        -- assume arg(1)=typeindicator, arg(2)=value
           return self~bsf.setFieldValueStrict(tmpName, args[1], args[2])
     end
end



-- say "BSF.CLS::unknown, name:" name "for" "'"self~objectname"'" "class:" self~class

/* 2004-02-22: can be longer than 250 chars which causes Object Rexx to bomb!
   code = 'call BSF "invoke", "' || self~objectname || '", "' || name || '"' makeArgString(args)
*/

   code = 'call BSF "invoke", "' || self~objectname || '", "' || name || '"'
   do i=1 to args~items
      a.i=choose(args[i]=.nil, .bsf~bsf.null_string, args[i])

      code=code ", a."i
   end

-- say "unknown, code: ["code"]... a.1="a.1

  signal on any         -- trap any exceptions
  interpret code        -- execute this dynamically created Rexx string
  if var("RESULT") then return bsf.checkResult(result)   -- check whether primitive type or Java object
                   else return .nil                      -- nothing came back, i.e. Java returned a true "null"
any: raise propagate  -- raise the exception in caller/invoker on the Rexx side




/* -------------------------------------------------------------
   the following methods implement the individual BSF-calls as
   methods, allowing for invoking them in the OO style with the
   help of messages;

   "registerBean" is not made available explicitly, it is
   implicitly part of the creation of a new proxy object (see
   "INIT" method above)

   "unregisterBean" is not made available explicitly, it is
   implicitly part of the destruction of objects (see the
   "UNINIT" methods above)

   "getStaticValue" is not made available to instances as they
   are able to get (and even set) public (static) fields via
   "getFieldValue" resp. "setFieldValue"
   ------------------------------------------------------------- */
--   instance methods

::METHOD bsf.class      -- method to return the BSF Java class object proxy of this Java object proxy
   PARSE VALUE self~objectname WITH java_path "@" .
   return .bsf~bsf.getJavaClassObjectProxy(java_path)


:: METHOD bsf.addEventListener  /* add event listener to the Java object*/
   use arg eventSetName, filter, text
   return BSF( "addEventListener", self~objectname, eventSetName, filter, text )

:: METHOD bsf.exit              /* exit Rexx                    */
   FORWARD MESSAGE ("BSF.EXIT") TO (self~class)

:: METHOD bsf.invoke
   parse arg methodName
   -- create UNKNOWN format (first argument is unknown method name)
   return self~unknown(methodName, arg(2, 'A'))


::METHOD bsf.invokeStrict
  USE ARG name          -- name of method to be invoked
  args=arg(2, 'A')

  -- 2004-02-22
  -- code = 'call BSF "invokeStrict", "' || self~objectname || '", "' || name || '"' makeArgString(args)

   code = 'call BSF "invokeStrict", "' || self~objectname || '", "' || name || '"'
   do i=1 to args~items
      a.i=choose(args[i]=.nil, .bsf~bsf.null_string, args[i])
      code=code ", a."i
   end

  -- say "invokeStrict, code=["code"]..."

  signal on any         -- trap any exceptions
  interpret code        -- execute this dynamically created Rexx string
  if var("RESULT") then return bsf.checkResult(result)   -- check whether primitive type or Java object
                   else return .nil                      -- nothing came back, i.e. Java returned a true "null"
any: raise propagate  -- raise the exception in caller/invoker on the Rexx side


:: METHOD bsf.getFieldValue
   parse arg fieldName
   return bsf.checkResult( bsf("getFieldValue", self~objectname, fieldName) )

:: METHOD bsf.getFieldValueStrict -- case-sensitive fieldName !
   parse arg fieldName
   return bsf.checkResult( bsf("getFieldValueStrict", self~objectname, fieldName) )


:: METHOD bsf.setFieldValue
   parse arg fieldName
   /* 2004-02-22
   code="res=bsf('setFieldValue', '" || self~objectname || "', '" || fieldName || ,
                 "'" makeArgString(arg(2, 'A')) ")"
   */

   code="res=bsf('setFieldValue', '" || self~objectname || "', '" || fieldName || "'" ,
                 ", arg(2) )"
  --  say "bsf.setFieldValue, code=["code"]..."

   signal on any         -- trap any exceptions
   interpret code
   return res

any: raise propagate  -- raise the exception in caller/invoker on the Rexx side

:: METHOD bsf.setFieldValueStrict
   parse arg fieldName
   /* 2004-02-22
   code="res=bsf('setFieldValueStrict', '" || self~objectname || "', '" || fieldName || ,
                 "'" makeArgString(arg(2, 'A')) ")"
   */

   code="res=bsf('setFieldValueStrict', '" || self~objectname || "', '" || fieldName || "'" ,
                 ", arg(2)" choose(arg()>2, ", arg(3)", "") ")"

   -- say "bsf.setFieldValueStrict, code=["code"]..."

   signal on any         -- trap any exceptions
   interpret code
   return res

any: raise propagate  -- raise the exception in caller/invoker on the Rexx side


:: METHOD bsf.getPropertyValue
   parse arg propertyName, index
   if arg(2, "Omitted") then index = "" -- no index given, set to empty string
   return bsf.checkResult( bsf("getPropertyValue", self~objectname, propertyName, index))


:: METHOD bsf.setPropertyValue
   parse arg propertyName, index
   if arg(2, "Omitted") then index = "" -- no index given, set to empty string
   /* 2004-02-22
   code="res=bsf('setPropertyValue', '" || self~objectname || "', '" || propertyName || "', '" ||,
                 index || "'" makeArgString(arg(3, 'A')) ")"
   */

   code="res=bsf('setPropertyValue', '" || self~objectname || "', '" || propertyName || "', '" ||,
                 index || "'"
   tmpArr=arg(3, 'Array')
   do i=1 to tmpArr~items
      a.i=choose(tmpArr[i]=.nil, .bsf~bsf.null_string, tmpArr[i])
      code=code ", a."i
   end
   code=code ")"

   signal on any         -- trap any exceptions
   interpret code
   return res

any: raise propagate  -- raise the exception in caller/invoker on the Rexx side


:: METHOD bsf.setPropertyValueStrict
   parse arg propertyName, index
   if arg(2, "Omitted") then index = "" -- no index given, set to empty string
   /* 2004-02-22
   code="res=bsf('setPropertyValueStrict', '" || self~objectname || "', '" || propertyName || "', '" ||,
                 index || "'" makeArgString(arg(3, 'A')) ")"
   */

   code="res=bsf('setPropertyValueStrict', '" || self~objectname || "', '" || propertyName || "', '" ||,
                 index || "'"
   tmpArr=arg(3, 'Array')
   do i=1 to tmpArr~items
      a.i=choose(tmpArr[i]=.nil, .bsf~bsf.null_string, tmpArr[i])
      code=code ", a."i
   end
   code=code ")"

   signal on any         -- trap any exceptions
   interpret code
   return res

any: raise propagate  -- raise the exception in caller/invoker on the Rexx side



--   class  methods

/* -------------------------------------------------------------
   class method shuts down the Java Virtual Machine (JVM), which started
   this instance of Rexx, this causes Rexx to be shut down as well; the
   Java side waits "waitTime" msecs before terminating the JVM
   ------------------------------------------------------------- */
::METHOD bsf.exit class
  use arg exitValue, waitTime

  if \datatype(exitValue, "Whole Number") then return BSF("exit")
  if \datatype(waitTime,  "W"           ) then return BSF("exit", exitValue)
  return bsf.checkResult(BSF("exit", exitValue, waitTime))

/* -------------------------------------------------------------
   class method to have Java sleep for us "sleepTime" seconds; resolution
   to a msec is calculated, although the Java side may deviate up to 15/100 secs
   (we are not talking realtime here ;)
   ------------------------------------------------------------- */
::METHOD bsf.sleep class
  use arg sleepTime

  return bsf.checkResult(BSF("sleep", sleepTime))

:: METHOD     bsf.lookupBean                class
   parse arg beanName
   return bsf.checkResult( bsf("lookupBean", beanName) )

:: METHOD     bsf.pollEventText             class
   parse arg timeOut
   if \datatype(timeOut, "Whole") then return bsf.checkResult( bsf("pollEventText"))
   return bsf.checkResult( bsf("pollEventText", timeout))    -- with timeout

:: METHOD     bsf.getStaticValue            class
   parse arg className, fieldName
   return bsf.checkResult( bsf("getStaticValue", className, fieldName))

:: METHOD     bsf.getStaticValueStrict      class
   parse arg className, fieldName
   return bsf.checkResult( bsf("getStaticValueStrict", className, fieldName))

:: METHOD     bsf.postEventText              class
   parse arg eventText
   return bsf("postEventText", eventText)

:: METHOD     bsf.wrapArray                  class
   parse arg beanName
   return bsf.checkResult( bsf("wrapArray", beanName) )

:: METHOD     bsf.createArray                class
   tmp="return bsf.checkResult(bsf('createArray'"
   do i=1 to arg()
      tmp=tmp ", """ || arg(i) || """"
   end
   tmp=tmp"))"

   signal on any         -- trap any exceptions
   interpret tmp     -- now carry out the Rexx statement
   return

any: raise propagate  -- raise the exception in caller/invoker on the Rexx side




:: METHOD     bsf.wrapEnumeration            class
   parse arg beanName
   return bsf.checkResult( bsf("wrapEnumeration", beanName) )

   -- set the string which from now on represents the Java "null" on both sides,
   -- the Rexx and the Java side
:: METHOD     bsf.setRexxNullString         class
   parse arg newValue
   .bsf~bsf.null_string = newValue                  -- set class attribute to the new value
   return bsf("setRexxNullString", newValue)    -- set the Java side to the new value



/* *************************************************************
   Class:       BSF_REFERENCE
   Purpose:     the proxy class for objects returned by the Java side;
                objects of this class, if destroyed, cause via its superclass
                to remove one from the appropriate instance counter of the
                BSF registry
   Restriction: Should only be used by classes/routines in this file

   ************************************************************* */
::CLASS BSF_REFERENCE subclass BSF

/* -------------------------------------------------------------
   method to construct an instance (a "bean"); as the object was returned
   from the Java side, just remember its name, but do not call BSF's init
   again! This proxy object merely allows for sending messages through it
   to the Java object as well. If it goes out of scope, it will be garbage
   collected by Object Rexx which in turn deregisters the object from the
   Java side, decrementing the reference counter there
   ------------------------------------------------------------- */
::METHOD init
   use arg beanName
   self~ObjectName=beanName



/* *************************************************************
   Class:       BSF_REFERENCE
   Purpose:     allows to wrap up Java objects as Object Rexx objects;
                if instances are destroyed, then no derigstration from
                the BSF registry takes place
   restriction: no check is done, whether the given beanName is really
                in the BSF-registry
   ************************************************************* */
::CLASS BSF_PROXY subclass BSF PUBLIC

/* -------------------------------------------------------------
   method to construct an instance (a "bean"); as the object was returned
   from the Java side, just remember its name, but do not call BSF's init
   again! This proxy object merely allows for sending messages through it
   to the Java object as well. If it goes out of scope, it will be garbage
   collected by Object Rexx which in turn unregisters the object from the
   Java side, decrementing the reference counter there
   ------------------------------------------------------------- */
::METHOD init
   use arg beanName
   self~ObjectName=beanName

-- say "---> BSF_PROXY.init  : bean=["||beanName||"]" -- 2003-01-29, ---rgf

   -- make sure that BSF's UNINIT method does not run (it'll deregister
   -- the bean from the BSF registry
::METHOD uninit
  NOP




/* *************************************************************
   Class:       BSF_ARRAY
   Purpose:     the proxy class for array objects returned by the Java side
   ************************************************************* */
::CLASS BSF_ARRAY_REFERENCE     subclass BSF

/* -------------------------------------------------------------
   method to construct an instance (a "bean"); as the object was returned
   from the Java side, just remember its name, but do not call BSF's init
   again! This proxy object merely allows for sending messages through it
   to the Java object as well. If it goes out of scope, it will be garbage
   collected by Object Rexx which in turn unregisters the object from the
   Java side, decrementing the reference counter there
   ------------------------------------------------------------- */
::METHOD init
   expose items dimension indexArray   -- set access to object variable
   use arg beanName, dimension  -- assign "dimension" directly to exposed object variable

   self~ObjectName=beanName
   items=-1                     -- set and indicate, that not calculated as of yet
      -- "indexArray" allows for indexing at's and put's with a Java int array
      -- use exact signature
   indexArray=bsf("invokeStrict", .bsf4rexx~array.class, "newInstance", "o", .bsf4rexx~int, "int", dimension)
   indexArray=.bsf_array_proxy~new(indexArray, 1, dimension)


::METHOD dimension
  expose dimension dimensions

  if arg(1, "Omitted") then return dimension    -- return number of dimensions

  if VAR( dimensions ) = .false then self~getData  -- no data available, get it
  return dimensions[arg(1)]                     -- return entries in this dimension



      -- get the needed information from ArrayWrapper, make it available with Object Rexx attributes
::METHOD getData private
  expose dimensions items

      -- get the array wrapper of this array
  aw=bsf('wrapArray', self~objectname)
  items=bsf('getFieldValue', aw, 'items')       -- get calculated 'items' value
  dimArr=bsf('getFieldValue', aw, 'sizeOfDimension') -- get the Java array containing the dimensions

  dimensions=.array~new                         -- create an Object Rexx array
  do i=1 to bsf('arrayLength', dimArr)          -- get length of 'dimensions' array
     dimensions[i]=bsf('arrayAt', dimArr, i-1)  -- use Java indexing (starting with '0')
  end


  -- Note: This array-proxy is "1"-based (not 0-based as Java), because Object Rexx arrays
  -- are "1", based, therefore subscript adjustments are necessary, before calling the Java side
  -- The indices can be given individually in comma-separated form or in form of a Rexx array
::METHOD at
   expose indexArray
   use arg arg1

      -- adjust Object Rexx 1-based indexing to Java 0-based indexing
   if arg1~class=.array then
   do i=1 to arg1~items
      -- call bsf "arrayPutStrict", indexArray, "int", arg1[i]-1, i-1
      call bsf "arrayPut", indexArray, arg1[i]-1, i-1
   end
   else -- individual enter the indices into the index-array
   do i=1 to arg()
      -- call bsf "arrayPutStrict", indexArray, "int", arg(i)-1, i-1
      call bsf "arrayPut", indexArray, arg(i)-1, i-1
   end
   return bsf.checkResult( bsf('arrayAt', self~objectName, indexArray))


  -- allow for referring to array elements as if it was an Object Rexx array
::METHOD "[]"
  FORWARD MESSAGE ("AT")        -- synonym for "AT", use its implementation



  -- Note: This array-proxy is "1"-based (not 0-based as Java), because Object Rexx arrays
  -- are "1", based, therefore subscript adjustments are necessary, before calling the Java side
  -- The indices can be given individually in comma-separated form or in form of a Rexx array
::METHOD put
   expose indexArray
   use arg value, tmpArr

   if value=.nil then value=.bsf~bsf.null_string -- indicate to the Java side a "null"

   if tmpArr~class<>.array then  -- if second argument is not an array, create one
      tmpArr=arg(2, "Array")     -- get rest of arguments as an array

   do i=1 to tmpArr~items        -- adjust indices to Java-style (0-based indices)
      -- call bsf "arrayPutStrict", indexArray, "int", tmpArr[i]-1, i-1
      call bsf "arrayPut", indexArray, tmpArr[i]-1, i-1
   end

   call bsf 'arrayPut', self~objectName, value, indexArray


  -- Note: This array-proxy is "1"-based (not 0-based as Java), because Object Rexx arrays
  -- are "1", based, therefore subscript adjustments are necessary, before calling the Java side
  -- The indices can be given individually in comma-separated form or in form of a Rexx array
::METHOD putStrict
   expose indexArray
   use arg type, value, tmpArr

   if value=.nil then value=.bsf~bsf.null_string -- indicate to the Java side a "null"

   if tmpArr~class<>.array then  -- if second argument is not an array, create one
      tmpArr=arg(2, "Array")     -- get rest of arguments as an array

   do i=1 to tmpArr~items        -- adjust indices to Java-style (0-based indices)
      -- call bsf "arrayPutStrict", indexArray, "int", tmpArr[i]-1, i-1
      call bsf "arrayPut", indexArray, tmpArr[i]-1, i-1
   end

   call bsf 'arrayPutStrict', self~objectName, type, value, indexArray



  -- allow for referring to array elements as if it was an Object Rexx array;
  -- using this syntax variant has the following implication: the valueType and the
  -- value itself must be given as *one* string value; the first word is the valueType
  -- the rest the value!    e.g.: xyz[1, 2, 3] = "String Hi, there!"
::METHOD "[]="
  FORWARD MESSAGE ("PUT") -- let the PUT-method do the work, 2003-05-10

  /* 2003-05-10, ---rgf
  -- synonym for "PUT", use its implementation
  expose dimension
  parse arg type value, .  -- for this variant extract the type from the rest of the value

  tmpArr=arg(2, "Array")   -- get the rest of the arguments in form of an array
  tmpArrItems=tmpArr~items
  newArr=.array~new
  newArr[1]=type           -- save type information
  newArr[2]=value          -- save value

  do i=3 to 9999           -- assign the index-arguments
     j=i-2
     if j>tmpArrItems then leave
     newArr[i]=tmpArr[j]
  end
  FORWARD MESSAGE ("PUT") ARGUMENTS (newArr) -- let the PUT-method do the work
  */


-- calculates (once) and returns number of individual elements in array
::METHOD items
  expose items  -- set access to object variable
  if items=-1 then self~getData     -- not yet set, get the data from the ArrayWrapper
  return items



-- this method creates a supplier method
::METHOD  SUPPLIER

  aw=bsf('wrapArray', self~objectname)
   -- wrap up the Supplier object, such that it will be kept in the BSF registry as long
   -- as the Object Rexx BSF proxy exists
  return bsf.checkResult( bsf('invoke', aw, 'supplier', .true) )



-- this method is needed by "DO ... OVER"
::METHOD MAKEARRAY

  aw=bsf('wrapArray', self~objectname)    -- wrap up the array
  arr=bsf('invoke', aw, 'makearray', .false, .false) -- get
  arr=bsf('arrayAt', arr, 0)              -- get the appropriate Java array

  rexxArr=.array~new
  do i=1 to bsf('arrayLength', arr)          -- get length of 'dimensions' array
     rexxArr[i]=bsf('arrayAt', arr, i-1)  -- use Java indexing (starting with '0')
  end
  return rexxArr



   -- wraps the given Java object and allows for using the array methods on it;
   -- if the reference is not used anymore, it will *not* cause the Java object to be de-registered
   -- used in the initialisation of "BSF_ARRAY" to get primitive int array (for the dimensions) without initialising them as standalone BSF-arrays
   --
::CLASS BSF_ARRAY_PROXY SUBCLASS BSF_ARRAY_REFERENCE
::METHOD init
  use arg beanName, dimension, items         -- name of Java object as registered in the BSF registry

  self~objectName=beanName
  self~dimension =dimension
  self~items     =items

::method dimension attribute
::method items     attribute

::METHOD uninit            -- intercept destructor message, don't remove BSF registry entry
  NOP




